@@ -1,14 +1,15 @@
 import os
 import re
 from timeit import default_timer
-from typing import Tuple
+from typing import Callable, Tuple
 
 from fastapi import FastAPI
-from prometheus_client import Histogram
 from starlette.requests import Request
 from starlette.responses import Response
 from starlette.routing import Match
 
+from prometheus_fastapi_instrumentator import metrics
+
 
 class PrometheusFastApiInstrumentator:
     def __init__(
@@ -18,10 +19,7 @@ class PrometheusFastApiInstrumentator:
         should_group_untemplated: bool = True,
         should_round_latency_decimals: bool = False,
         should_respect_env_var: bool = False,
-        excluded_handlers: list = ["/metrics"],
-        buckets: tuple = Histogram.DEFAULT_BUCKETS,
-        metric_name: str = "http_request_duration_seconds",
-        label_names: tuple = ("method", "handler", "status",),
+        excluded_handlers: list = [],
         round_latency_decimals: int = 4,
         env_var_name: str = "ENABLE_METRICS",
     ):
@@ -44,23 +42,12 @@ class PrometheusFastApiInstrumentator:
             FastAPI app that is used by multiple distinct apps. The apps only 
             have to set the variable to be instrumented.
 
-        :param excluded_handlers: Handlers that should be ignored. List of 
-            strings is turned into regex patterns.
-
-        :param buckets: Buckets for the histogram. Defaults to Prometheus default.
-
-        :param metric_name: Name of the latency metric.
-
-        :param label_names: Names of the three labels used for the metric. Does 
-            not influence the label values. Defaults to ("method", "handler", 
-            "status",).
-
         :param round_latency_decimals: Number of decimals latencies should be 
             rounded to. Ignored unless `should_round_latency_decimals` is `True`.
 
         :param env_var_name: Any valid os environment variable name that 
             will be checked for existence before instrumentation. Ignored 
-            unless `should_respect_env_var_existence` is `True`.
+            unless `should_respect_env_var` is `True`.
         """
 
         self.should_group_status_codes = should_group_status_codes
@@ -70,7 +57,6 @@ class PrometheusFastApiInstrumentator:
         self.should_respect_env_var = should_respect_env_var
 
         self.round_latency_decimals = round_latency_decimals
-        self.label_names = label_names
         self.env_var_name = env_var_name
 
         if excluded_handlers:
@@ -78,17 +64,7 @@ class PrometheusFastApiInstrumentator:
         else:
             self.excluded_handlers = []
 
-        if buckets[-1] == float("inf"):
-            self.buckets = buckets
-        else:
-            self.buckets = buckets + (float("inf"),)
-
-        self.histogram = Histogram(
-            name=metric_name,
-            documentation="Duration of HTTP requests in seconds",
-            labelnames=self.label_names,
-            buckets=self.buckets,
-        )
+        self.instrumentations = []
 
     def instrument(self, app: FastAPI) -> "self":
         """Performs the instrumentation by adding middleware and endpoint.
@@ -103,38 +79,46 @@ class PrometheusFastApiInstrumentator:
         ):
             return self
 
+        if len(self.instrumentations) == 0:
+            self.instrumentations.append(metrics.http_request_duration_seconds())
+
         @app.middleware("http")
         async def dispatch_middleware(request: Request, call_next) -> Response:
             start_time = default_timer()
 
-            method = request.method
-
-            handler, is_templated = self._get_handler(request)
-
             try:
                 response = None
                 response = await call_next(request)
+                status = str(response.status_code)
             except Exception as e:
                 if response is None:
-                    status = 500
+                    status = "500"
                 raise e from None
-            else:
-                status = response.status_code
+            finally:
+                handler, is_templated = self._get_handler(request)
 
-            if self._is_handler_excluded(handler, is_templated):
-                return response
+                if not self._is_handler_excluded(handler, is_templated):
+                    duration = max(default_timer() - start_time, 0)
 
-            if is_templated is False and self.should_group_untemplated:
-                handler = "none"
+                    if self.should_round_latency_decimals:
+                        duration = round(duration, self.round_latency_decimals)
 
-            duration = max(default_timer() - start_time, 0)
+                    if is_templated is False and self.should_group_untemplated:
+                        handler = "none"
 
-            if self.should_round_latency_decimals:
-                duration = round(duration, self.round_latency_decimals)
+                    if self.should_group_status_codes:
+                        status = status[0] + "xx"
 
-            self.histogram.labels(
-                *self._create_label_tuple(method, handler, status)
-            ).observe(duration)
+                    info = metrics.Info(
+                        request=request,
+                        response=response,
+                        modified_handler=handler,
+                        modified_status=status,
+                        modified_duration=duration,
+                    )
+
+                    for instrumentation in self.instrumentations:
+                        instrumentation(info)
 
             return response
 
@@ -161,9 +145,13 @@ class PrometheusFastApiInstrumentator:
         ):
             return self
 
-        from prometheus_client import (CONTENT_TYPE_LATEST, REGISTRY,
-                                       CollectorRegistry, generate_latest,
-                                       multiprocess)
+        from prometheus_client import (
+            CONTENT_TYPE_LATEST,
+            REGISTRY,
+            CollectorRegistry,
+            generate_latest,
+            multiprocess,
+        )
 
         if "prometheus_multiproc_dir" in os.environ:
             pmd = os.environ["prometheus_multiproc_dir"]
@@ -183,28 +171,9 @@ class PrometheusFastApiInstrumentator:
 
         return self
 
-    def _create_label_tuple(
-        self, method: str, handler: str, code: int
-    ) -> Tuple[str, str, str]:
-        """Processes label values based on config.
-
-        :param method: Method used for the request, for example `GET`.
-        :param handler: Identifier for entity that handled / should handle the 
-            request.
-        :param code: Status code of the response. If error / exception occured 
-            this should be `500`.
-        :return: Processed values.
-        """
-
-        code = str(code)
-
-        if self.should_group_status_codes:
-            code = code[0] + "xx"
-        return (
-            method,
-            handler,
-            code,
-        )
+    def add(self, instrumentation_function: Callable[[metrics.Info], None]) -> "self":
+        self.instrumentations.append(instrumentation_function)
+        return self
 
     def _get_handler(self, request: Request) -> Tuple[str, bool]:
         """Extracts either template or (if no template) path."""
